from ..utils import load_config_key
from ..custom_logging import log

def can_agent_stream(agent_name: str):

    log.debug(f"agent_type: {agent_name} checking streaming...")
    endpoints_config = load_config_key(agent_name, "dummy_value", kind="agentConfig")
    post_endpoints = endpoints_config['post']
    
    return 'stream' in post_endpoints

    
    


# process_azure_blob_event.py
from ..custom_logging import log

def process_azure_blob_event(events: list) -> tuple:
    """
    Extracts message data and metadata from an Azure Blob Storage event.

    Args:
        events (list): The list of Azure Event Grid event data.

    Returns:
        tuple: A tuple containing the blob URL, attributes as metadata, and the vector name.
        
    Example of Event Grid schema:
    ```
    {
        "topic": "/subscriptions/subscription-id/resourceGroups/resource-group/providers/Microsoft.Storage/storageAccounts/storage-account",
        "subject": "/blobServices/default/containers/container/blobs/blob",
        "eventType": "Microsoft.Storage.BlobCreated",
        "eventTime": "2021-01-01T12:34:56.789Z",
        "id": "event-id",
        "data": {
            "api": "PutBlob",
            "clientRequestId": "client-request-id",
            "requestId": "request-id",
            "eTag": "etag",
            "contentType": "application/octet-stream",
            "contentLength": 524288,
            "blobType": "BlockBlob",
            "url": "https://storage-account.blob.core.windows.net/container/blob",
            "sequencer": "0000000000000000000000000",
            "storageDiagnostics": {
                "batchId": "batch-id"
            }
        },
        "dataVersion": "",
        "metadataVersion": "1"
    }
    ```
    """
    storage_blob_created_event = "Microsoft.Storage.BlobCreated"
    
    for event in events:
        event_type = event['eventType']
        data = event['data']

        if event_type == storage_blob_created_event:
            blob_url = data['url']
            event_time = event['eventTime']
            event_id = event['id']
            subject = event['subject']
            attributes = {
                'event_type': event_type,
                'event_time': event_time,
                'event_id': event_id,
                'subject': subject,
                'url': blob_url
            }

            vector_name = subject.split('/')[4]  # Extracting the container name
            
            log.info(f"Process Azure Blob Event was triggered by eventId {event_id} at {event_time}")
            log.debug(f"Process Azure Blob Event data: {blob_url}")
            
            # Check for a valid Azure Blob Storage event type
            if event_type == "Microsoft.Storage.BlobCreated":
                log.info(f"Got valid event from Azure Blob Storage: {blob_url}")
            
            return blob_url, attributes, vector_name

    return None, None, None


#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from ..custom_logging import log

import pathlib
import tempfile
import os
import re
import json

try:
    from google.cloud import storage
except ImportError:
    storage = None

try:
    from azure.storage.blob import BlobServiceClient
except ImportError:
    BlobServiceClient = None

from ..langchain_types import Document

from .splitter import chunk_doc_to_docs
from .pdfs import split_pdf_to_pages
from .publish import publish_if_urls
from . import loaders

from ..utils.parsers import extract_urls
from ..gcs.add_file import add_file_to_gcs, get_pdf_split_file_name
from ..azure.blobs import extract_blob_parts
from ..azure.auth import azure_auth

def handle_gcs_message(message_data: str, metadata: dict, vector_name: str):

    if not storage:
        log.warning("No GCS storage client")
        return None, None
    # Process message from Google Cloud Storage
    log.debug("Detected gs://")
    bucket_name, file_name = message_data[5:].split("/", 1)

    # Create a client
    storage_client = storage.Client()

    # Download the file from GCS
    bucket = storage_client.get_bucket(bucket_name)
    blob = bucket.blob(file_name)

    file_name=pathlib.Path(file_name)

    with tempfile.TemporaryDirectory() as temp_dir:
        tmp_file_path = os.path.join(temp_dir, file_name.name)
        blob.download_to_filename(tmp_file_path)

        if file_name.suffix.lower() == ".pdf":
            pages = split_pdf_to_pages(tmp_file_path, temp_dir)
            if not metadata.get("source"):
                metadata["source"] = str(file_name)
            if len(pages) > 1: # we send it back to GCS to parrallise the imports
                log.info(f"Got back {len(pages)} pages for file {tmp_file_path}")
                for pp in pages:
                    pp_basename = os.path.basename(pp)
                    # file_name/pdf_parts/file_name_1.pdf
                    bucket_path = get_pdf_split_file_name(file_name, part_name=pp_basename)
                    gs_file = add_file_to_gcs(pp, 
                                              vector_name=vector_name, 
                                              bucket_name=bucket_name, 
                                              metadata=metadata,
                                              bucket_filepath=bucket_path)
                    log.info(f"{gs_file} is now in bucket {bucket_name}")
                log.info(f"Sent split pages for {file_name.name} back to GCS to parrallise the imports")
                return None, None
        else:
            # just original temp file
            pages = [tmp_file_path]

        the_metadata = {
            "type": "file_load_gcs",
            "bucket_name": bucket_name
        }
        
        if metadata.get("source") is None:
            the_metadata["source"] = str(file_name)

        metadata.update(the_metadata)

        docs = []
        for page in pages:
            log.info(f"Sending file {page} to loaders.read_file_to_documents {metadata}")
            docs2 = loaders.read_file_to_documents(page, metadata=metadata)
            if docs2 is None:
                log.warning(f"loaders.read_file_to_documents docs2 failed to load file {metadata}")
            docs.extend(docs2)

        if docs is None:
            log.warning(f"loaders.read_file_to_documents docs failed to load file {metadata}")
            return None, metadata
        else:
            chunks = chunk_doc_to_docs(docs, file_name.suffix, vector_name=vector_name)

        return chunks, metadata

def handle_google_drive_message(message_data: str, metadata: dict, vector_name: str):
    # Process message from Google Drive
    log.info("Got google drive URL")
    urls = extract_urls(message_data)

    docs = []
    for url in urls:
        metadata["source"] = url
        metadata["url"] = url
        metadata["type"] = "url_load"
        doc = loaders.read_gdrive_to_document(url, metadata=metadata)
        if doc is None:
            log.info("Could not load any Google Drive docs")
        else:
            docs.extend(doc)

    chunks = chunk_doc_to_docs(docs, vector_name=vector_name)

    return chunks, metadata

def handle_github_message(message_data: str, metadata: dict, vector_name: str):
    # Process message from GitHub
    log.info("Got GitHub URL")
    urls = extract_urls(message_data)

    branch="main"
    if "branch:" in message_data:
        match = re.search(r'branch:(\w+)', message_data)
        if match:
            branch = match.group(1)
    
    log.info(f"Using branch: {branch}")

    docs = []
    for url in urls:
        metadata["source"] = url
        metadata["url"] = url
        metadata["type"] = "url_load"
        doc = loaders.read_git_repo(url, branch=branch, metadata=metadata)
        if doc is None:
            log.info("Could not load GitHub files")
        else:
            docs.extend(doc)
    
    chunks = chunk_doc_to_docs(docs, vector_name=vector_name)
    
    return chunks, metadata

def handle_http_message(message_data: str, metadata: dict, vector_name:str):
    # Process message from a generic HTTP URL
    log.info(f"Got http message: {message_data}")

    # just in case, extract the URL again
    urls = extract_urls(message_data)

    docs = []
    for url in urls:
        metadata["source"] = url
        metadata["url"] = url
        metadata["type"] = "url_load"
        doc = loaders.read_url_to_document(url, metadata=metadata)
        docs.extend(doc)

    chunks = chunk_doc_to_docs(docs, vector_name=vector_name)

    return chunks, metadata

def handle_json_content_message(message_data: dict, metadata: dict, vector_name: str):
    log.info("No tailored message_data detected, processing message json")
    # Process message containing direct JSON content
    try:
        the_json = json.loads(message_data)
    except Exception as e:
        log.error(f"Could not load message {message_data} as JSON - {str(e)}")
        return None, {"metadata": f"Could not load message as JSON - {str(e)}"}
    
    the_metadata = the_json.get("metadata", {})
    metadata.update(the_metadata)
    the_content = the_json.get("page_content", None)

    if metadata.get("source", None) is not None:
        metadata["source"] = "No source embedded"

    if the_content is None:
        log.info("No content found")
        return None, {"metadata": "No content found in 'page_content' JSON field"}
    
    docs = [Document(page_content=the_content, metadata=metadata)]

    publish_if_urls(the_content, vector_name)

    chunks = chunk_doc_to_docs(docs, vector_name=vector_name)

    return chunks, metadata


def handle_azure_blob(message_data: str, metadata: dict, vector_name: str):
    """
    Processes a message from Azure Blob storage, downloads the file, processes it,
    and returns chunks and metadata.

    Args:
        message_data (str): URL of the Azure blob.
        metadata (dict): Metadata associated with the file.
        vector_name (str): Vector name for processing.

    Returns:
        chunks (list): List of document chunks.
        metadata (dict): Updated metadata.
    """

    if BlobServiceClient is None:
        raise ImportError("BlobServiceClient is not installed - install via pip install sunholo[azure]")

    account_name, container_name, blob_name = extract_blob_parts(message_data)

    credential = azure_auth()
    if credential is None:
        log.error("BlobServiceClient could not find auth credentials")
        return None, None
    
    # Create a BlobServiceClient
    blob_service_client = BlobServiceClient(
        account_url=f"https://{account_name}.blob.core.windows.net",
        credential=credential)

    # Get the blob client
    blob_client = blob_service_client.get_blob_client(container=container_name, blob=blob_name)

    file_name = pathlib.Path(blob_name)

    with tempfile.TemporaryDirectory() as temp_dir:
        tmp_file_path = os.path.join(temp_dir, file_name.name)
        with open(tmp_file_path, "wb") as file:
            download_stream = blob_client.download_blob()
            file.write(download_stream.readall())

        if file_name.suffix.lower() == ".pdf":
            pages = split_pdf_to_pages(tmp_file_path, temp_dir)
            if not metadata.get("source"):
                metadata["source"] = str(file_name)
            if len(pages) > 1:
                log.info(f"Got back {len(pages)} pages for file {tmp_file_path}")
                for pp in pages:
                    pp_basename = os.path.basename(pp)
                    # file_name/pdf_parts/file_name_1.pdf
                    azure_blob_path = f"{file_name.stem}_parts/{pp_basename}"

                    # Upload split pages back to Azure Blob storage
                    container_client = blob_service_client.get_container_client(container=container_name)
                    with open(file=pp, mode="rb") as page_file:
                        blob_client = container_client.upload_blob(name=azure_blob_path, data=page_file, overwrite=True)

                    log.info(f"{azure_blob_path} is now in container {container_name}")
                log.info(f"Sent split pages for {file_name.name} back to Azure Blob to parallelize the imports")
                return None, None
        else:
            # just original temp file
            pages = [tmp_file_path]

        the_metadata = {
            "type": "file_load_azure_blob",
            "container_name": container_name
        }

        if metadata.get("source") is None:
            the_metadata["source"] = str(file_name)

        metadata.update(the_metadata)

        docs = []
        for page in pages:
            log.info(f"Sending file {page} to loaders.read_file_to_documents {metadata}")
            docs2 = loaders.read_file_to_documents(page, metadata=metadata)
            if docs2 is None:
                log.warning(f"loaders.read_file_to_documents docs2 failed to load file {metadata}")
            docs.extend(docs2)

        if docs is None:
            log.warning(f"loaders.read_file_to_documents docs failed to load file {metadata}")
            return None, metadata
        else:
            chunks = chunk_doc_to_docs(docs, file_name.suffix, vector_name=vector_name)

        return chunks, metadata


#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import json, os
import requests
from ..custom_logging import log



def generate_discord_output(bot_output):
    source_documents = []
    if bot_output.get('source_documents', None) is not None:
        source_documents = []
        for doc in bot_output['source_documents']:
            metadata = doc.get("metadata",{})
            filtered_metadata = {}
            if metadata.get("source", None) is not None:
                filtered_metadata["source"] = metadata["source"]
            if metadata.get("type", None) is not None:
                filtered_metadata["type"] = metadata["type"]
            source_doc = {
                'page_content': doc["page_content"],
                'metadata': filtered_metadata
            }
            source_documents.append(source_doc)

    return {
        'result': bot_output.get('answer', "No answer available"),
        'source_documents': source_documents
    }

def discord_webhook(message_data):
    webhook_url = os.getenv('DISCORD_URL', None)  # replace with your webhook url
    if webhook_url is None:
        return None
    
    log.info(f'webhook url: {webhook_url}')

    # If the message_data is not a dict, wrap it in a dict.
    if not isinstance(message_data, dict):
        message_data = {'content': message_data}
    else:
        # if it is a dict, turn it into a string
        message_data = {'content': json.dumps(message_data)}
        #TODO parse out message_data into other discord webhook objects like embed
        # https://birdie0.github.io/discord-webhooks-guide/discord_webhook.html
    
    data = message_data

    log.info(f'Sending discord this data: {data}')
    response = requests.post(webhook_url, json=data,
                            headers={'Content-Type': 'application/json'})
    log.debug(f'Sent data to discord: {response}')
    
    return response



from sunholo.utils import ConfigManager
from sunholo.vertex import (
    init_genai,
)

from tools.your_agent import get_quarto, quarto_content, QuartoProcessor

from my_log import log

init_genai()

# kwargs supports - image_uri, mime
def vac_stream(question: str, vector_name:str, chat_history=[], callback=None, **kwargs):
    
    config=ConfigManager(vector_name)
    processor = QuartoProcessor(config)

    orchestrator = get_quarto(config, processor)
    if not orchestrator:
        msg = f"No quarto model could be configured for {vector_name}"
        log.error(msg)
        callback.on_llm_end(response=msg)
        return {"answer": msg}

    chat = orchestrator.start_chat()

    guardrail = 0
    guardrail_max = kwargs.get('max_steps', 10)
    big_text = ""
    usage_metadata = None
    functions_called = []
    result=None
    last_responses=None
    while guardrail < guardrail_max:

        content = quarto_content(question, chat_history)
        log.info(f"# Loop [{guardrail}] - {content=}")
        response = chat.send_message(content, stream=True)
        this_text = "" # reset for this loop
        log.debug(f"[{guardrail}] {response}")

        for chunk in response:
            try:
                log.debug(f"[{guardrail}] {chunk=}")
                # Check if 'text' is an attribute of chunk and if it's a string
                if hasattr(chunk, 'text') and isinstance(chunk.text, str):
                    token = chunk.text
                else:
                    function_names = []
                    try:
                        for part in chunk.candidates[0].content.parts:
                            if fn := part.function_call:
                                params = {key: val for key, val in fn.args.items()}
                                func_args = ",".join(f"{key}={value}" for key, value in params.items())
                                log.info(f"Found function call: {fn.name}({func_args})")
                                function_names.append(f"{fn.name}({func_args})")
                                functions_called.append(f"{fn.name}({func_args})")
                    except Exception as err:
                        log.warning(f"{str(err)}")

                    token = ""  # Handle the case where 'text' is not available
                    
                    if processor.last_api_requests_and_responses:
                        if processor.last_api_requests_and_responses != last_responses:
                            last_responses = processor.last_api_requests_and_responses
                        for last_response in last_responses:
                            result=None # reset for this function response
                            if last_response:
                                log.info(f"[{guardrail}] {last_response=}")
                                
                                # Convert the last_response to a string by extracting relevant information
                                function_name = last_response[0]
                                arguments = last_response[1]
                                result = last_response[2]
                                func_args = ",".join(f"{key}={value}" for key, value in arguments.items())

                                if f"{function_name}({func_args})" not in function_names:
                                    log.warning(f"skipping {function_name}({func_args}) as not in execution list")
                                    continue

                                token = f"\n## Loop [{guardrail}] Function call: {function_name}({func_args}):\n"

                                if function_name == "decide_to_go_on":
                                    token += f"# go_on={result}\n"
                                else:
                                    log.info("Adding result for: {function_name}")
                                    token += result

                callback.on_llm_new_token(token=token)
                big_text += token
                this_text += token
                
                if not usage_metadata:
                    chunk_metadata = chunk.usage_metadata
                    usage_metadata = {
                        "prompt_token_count": chunk_metadata.prompt_token_count,
                        "candidates_token_count": chunk_metadata.candidates_token_count,
                        "total_token_count": chunk_metadata.total_token_count,
                    }

            except ValueError as err:
                callback.on_llm_new_token(token=str(err))
        
        # change response to one with executed functions
        response = processor.process_funcs(response)

        if this_text:
            chat_history.append(("<waiting for ai>", this_text))
            log.info(f"[{guardrail}] Updated chat_history: {chat_history}")

        go_on_check = processor.check_function_result("decide_to_go_on", False)
        if go_on_check:
            log.info("Breaking agent loop")
            break
        
        guardrail += 1
        if guardrail > guardrail_max:
            log.warning("Guardrail kicked in, more than 10 loops")
            break

    callback.on_llm_end(response=big_text)
    log.info(f"orchestrator.response: {big_text}")

    metadata = {
        "question:": question,
        "chat_history": chat_history,
        "usage_metadata": usage_metadata,
        "functions_called": functions_called
    }

    return {"answer": big_text or "No answer was given", "metadata": metadata}


def vac(question: str, vector_name: str, chat_history=[], **kwargs):
    # Create a callback that does nothing for streaming if you don't want intermediate outputs
    class NoOpCallback:
        def on_llm_new_token(self, token):
            pass
        def on_llm_end(self, response):
            pass

    # Use the NoOpCallback for non-streaming behavior
    callback = NoOpCallback()

    # Pass all arguments to vac_stream and use the final return
    result = vac_stream(
        question=question, 
        vector_name=vector_name, 
        chat_history=chat_history, 
        callback=callback, 
        **kwargs
    )

    return result





from .streaming import start_streaming_chat, generate_proxy_stream, generate_proxy_stream_async, start_streaming_chat_async
from .langserve import parse_langserve_token, parse_langserve_token_async
from .stream_lookup import can_agent_stream


import json

try:
    from ..cli.sun_rich import console
except ImportError:
    console = None

from ..custom_logging import log

# Make sure to adjust the relative import path if needed
from .discovery_engine_client import DiscoveryEngineClient  

def discovery_engine_command(args):
    """
    Handles the `discovery-engine` command and its subcommands.
    """
    if args.subcommand == 'create-datastore':
        create_datastore_command(args)
    elif args.subcommand == 'import-documents':
        import_documents_command(args)
    elif args.subcommand == 'import-documents-with-metadata':
        import_documents_with_metadata_command(args)
    elif args.subcommand == 'import-document-with-metadata':
        import_document_with_metadata_command(args)
    elif args.subcommand == 'search':
        search_command(args)
    elif args.subcommand == 'search-by-id-and-or-date':
        search_by_id_and_or_date_command(args)
    else:
        console.print(f"[bold red]Unknown Discovery Engine subcommand: {args.subcommand}[/bold red]")

def create_datastore_command(args):
    """
    Handles the `discovery-engine create-datastore` subcommand.
    """
    try:
        client = DiscoveryEngineClient(
            project_id=args.project,
            data_store_id=args.data_store_id,
            location=args.location
        )
        operation_name = client.create_data_store(
            type=args.type,
            chunk_size=args.chunk_size,
            collection=args.collection
        )
        console.print(f"[bold green]Datastore creation initiated. Operation name: {operation_name}[/bold green]")
    except Exception as e:
        console.print(f"[bold red]Error creating datastore: {e}[/bold red]")

def import_documents_command(args):
    """
    Handles the `discovery-engine import-documents` subcommand.
    """
    try:
        client = DiscoveryEngineClient(
            project_id=args.project,
            data_store_id=args.data_store_id,
            location=args.location
        )
        operation_name = client.import_documents(
            gcs_uri=args.gcs_uri,
            data_schema=args.data_schema,
            branch=args.branch,
            bigquery_dataset=args.bigquery_dataset,
            bigquery_table=args.bigquery_table,
            bigquery_project_id=args.bigquery_project_id
        )
        console.print(f"[bold green]Document import initiated. Operation name: {operation_name}[/bold green]")
    except Exception as e:
        console.print(f"[bold red]Error importing documents: {e}[/bold red]")

def import_documents_with_metadata_command(args):
    """
    Handles the `discovery-engine import-documents-with-metadata` subcommand.
    """
    try:
        client = DiscoveryEngineClient(
            project_id=args.project,
            data_store_id=args.data_store_id,
            location=args.location
        )
        operation_name = client.import_documents_with_metadata(
            gcs_uri=args.gcs_uri,
            data_schema=args.data_schema,
            branch=args.branch
        )
        console.print(f"[bold green]Document import with metadata initiated. Operation name: {operation_name}[/bold green]")
    except Exception as e:
        console.print(f"[bold red]Error importing documents with metadata: {e}[/bold red]")

def import_document_with_metadata_command(args):
    """
    Handles the `discovery-engine import-document-with-metadata` subcommand.
    """
    try:
        # Load metadata from JSON file or string
        if args.metadata_file:
            with open(args.metadata_file, 'r') as f:
                metadata = json.load(f)
        elif args.metadata_string:
            metadata = json.loads(args.metadata_string)
        else:
            console.print("[bold red]Error: Must provide either --metadata-file or --metadata-string[/bold red]")
            return

        client = DiscoveryEngineClient(
            project_id=args.project,
            data_store_id=args.data_store_id,
            location=args.location
        )
        operation_name = client.import_document_with_metadata(
            gcs_uri=args.gcs_uri,
            metadata=metadata,
            branch=args.branch
        )
        console.print(f"[bold green]Document import with metadata initiated. Operation name: {operation_name}[/bold green]")
    except Exception as e:
        console.print(f"[bold red]Error importing document with metadata: {e}[/bold red]")

def search_command(args):
    """
    Handles the `discovery-engine search` subcommand.
    """
    try:
        client = DiscoveryEngineClient(
            project_id=args.project,
            data_store_id=args.data_store_id,
            location=args.location
        )
        results = client.get_chunks(
            query=args.query,
            num_previous_chunks=args.num_previous_chunks,
            num_next_chunks=args.num_next_chunks,
            page_size=args.page_size,
            parse_chunks_to_string=args.parse_chunks_to_string,
            serving_config=args.serving_config,
            data_store_ids=args.data_store_ids
        )

        if args.parse_chunks_to_string:
            console.print(results)  # Print the combined string
        else:
            # Process and print the results (assuming it's a SearchResponse object)
            for result in results.results:
                for chunk in result.document.chunks:
                    console.print(f"Chunk: {chunk.snippet}, document name: {chunk.document_name}")
    except Exception as e:
        console.print(f"[bold red]Error searching: {e}[/bold red]")

def search_by_id_and_or_date_command(args):
    """
    Handles the `discovery-engine search-by-id-and-or-date` subcommand.
    """
    try:
        client = DiscoveryEngineClient(
            project_id=args.project,
            data_store_id=args.data_store_id,
            location=args.location
        )
        results = client.search_by_objectId_and_or_date(
            query=args.query,
            objectId=args.object_id,
            date=args.date,
            num_previous_chunks=args.num_previous_chunks,
            num_next_chunks=args.num_next_chunks,
            page_size=args.page_size,
            parse_chunks_to_string=args.parse_chunks_to_string,
            serving_config=args.serving_config,
            data_store_ids=args.data_store_ids
        )

        if args.parse_chunks_to_string:
            console.print(results)  # Print the combined string
        else:
            # Process and print the results (assuming it's a SearchResponse object)
            for result in results.results:
                for chunk in result.document.chunks:
                    console.print(f"Chunk: {chunk.snippet}, document name: {chunk.document_name}")
    except Exception as e:
        console.print(f"[bold red]Error searching by ID and/or date: {e}[/bold red]")

def setup_discovery_engine_subparser(subparsers):
    """
    Sets up the `discovery-engine` subparser and its subcommands.
    """
    discovery_engine_parser = subparsers.add_parser('discovery-engine', help='Interact with Google Cloud Discovery Engine')
    discovery_engine_subparsers = discovery_engine_parser.add_subparsers(dest='subcommand', required=True)

    # Create Datastore subcommand
    create_datastore_parser = discovery_engine_subparsers.add_parser('create-datastore', help='Create a new Discovery Engine datastore')
    create_datastore_parser.add_argument('--data-store-id', required=True, help='The ID of the datastore')
    create_datastore_parser.add_argument('--type', choices=['chunk'], default='chunk', help='The type of datastore to create')
    create_datastore_parser.add_argument('--chunk-size', type=int, default=500, help='The size of the chunks for documents (if applicable)')
    create_datastore_parser.add_argument('--collection', default='default_collection', help='The collection to create the datastore in')
    create_datastore_parser.set_defaults(func=discovery_engine_command)

    # Import Documents subcommand
    import_documents_parser = discovery_engine_subparsers.add_parser('import-documents', help='Import documents into a Discovery Engine datastore')
    import_documents_parser.add_argument('--gcs-uri', required=True, help='The GCS URI of the documents to import')
    import_documents_parser.add_argument('--data-schema', default='content', help='The schema of the data to import')
    import_documents_parser.add_argument('--branch', default='default_branch', help='The branch to import the documents into')
    import_documents_parser.add_argument('--bigquery-dataset', help='The BigQuery dataset ID (if applicable)')
    import_documents_parser.add_argument('--bigquery-table', help='The BigQuery table ID (if applicable)')
    import_documents_parser.add_argument('--bigquery-project-id', help='The project ID of the BigQuery dataset (if applicable)')
    import_documents_parser.set_defaults(func=discovery_engine_command)

    # Import Documents with Metadata subcommand
    import_documents_with_metadata_parser = discovery_engine_subparsers.add_parser('import-documents-with-metadata', help='Import documents with metadata into a Discovery Engine datastore')
    import_documents_with_metadata_parser.add_argument('--gcs-uri', required=True, help='The GCS URI of the documents to import (JSONL format with metadata)')
    import_documents_with_metadata_parser.add_argument('--data-schema', default='content', help='The schema of the data to import')
    import_documents_with_metadata_parser.add_argument('--branch', default='default_branch', help='The branch to import the documents into')
    import_documents_with_metadata_parser.set_defaults(func=discovery_engine_command)

    # Import Document with Metadata subcommand
    import_document_with_metadata_parser = discovery_engine_subparsers.add_parser('import-document-with-metadata', help='Import a single document with metadata into a Discovery Engine datastore')
    import_document_with_metadata_parser.add_argument('--gcs-uri', required=True, help='The GCS URI of the document to import')
    import_document_with_metadata_parser.add_argument('--metadata-file', help='The path to a JSON file containing the metadata')
    import_document_with_metadata_parser.add_argument('--metadata-string', help='A JSON string containing the metadata')
    import_document_with_metadata_parser.add_argument('--branch', default='default_branch', help='The branch to import the document into')
    import_document_with_metadata_parser.set_defaults(func=discovery_engine_command)

    # Search subcommand
    search_parser = discovery_engine_subparsers.add_parser('search', help='Search a Discovery Engine datastore')
    search_parser.add_argument('--query', required=True, help='The search query')
    search_parser.add_argument('--data-store-id', required=True, help='Data store ID to search')
    search_parser.add_argument('--page-size', type=int, default=10, help='The maximum number of results to return per page')
    search_parser.add_argument('--parse-chunks-to-string', action='store_true', help='Combine chunks into a single string')
    search_parser.add_argument('--serving-config', default='default_serving_config', help='The serving configuration to use')

    search_parser.set_defaults(func=discovery_engine_command)

    # Search by ID and/or Date subcommand
    search_by_id_and_or_date_parser = discovery_engine_subparsers.add_parser('search-by-id-and-or-date', help='Search a Discovery Engine datastore by object ID and/or date')
    search_by_id_and_or_date_parser.add_argument('--query', required=True, help='The search query')
    search_by_id_and_or_date_parser.add_argument('--object-id', help='The exact object ID to filter by')
    search_by_id_and_or_date_parser.add_argument('--date', help='The date to filter by (YYYY-MM-DD)')
    search_by_id_and_or_date_parser.add_argument('--num-previous-chunks', type=int, default=3, help='Number of previous chunks to return for context')
    search_by_id_and_or_date_parser.add_argument('--num-next-chunks', type=int, default=3, help='Number of next chunks to return for context')
    search_by_id_and_or_date_parser.add_argument('--page-size', type=int, default=10, help='The maximum number of results to return per page')
    search_by_id_and_or_date_parser.add_argument('--parse-chunks-to-string', action='store_true', help='Combine chunks into a single string')
    search_by_id_and_or_date_parser.add_argument('--serving-config', default='default_serving_config', help='The serving configuration to use')
    search_by_id_and_or_date_parser.add_argument('--data-store-ids', nargs='+', help='List of data store IDs to search (optional)')
    search_by_id_and_or_date_parser.set_defaults(func=discovery_engine_command)



import datetime
from ..custom_logging import log

def audit_metadata(metadata, chunk_length=None):
    
    if 'eventTime' not in metadata:
        metadata['eventTime'] = datetime.datetime.now().isoformat(timespec='microseconds') + "Z"
    metadata['eventtime'] = metadata['eventTime']

    if 'source' not in metadata:
        if 'objectId' in metadata:
            metadata['source'] = metadata['objectId']
        elif 'url' in metadata:
            metadata['source'] = metadata['url']
        else:
            log.warning(f"No source found in metadata: {metadata}")
    
    if 'original_source' not in metadata:
        metadata['original_source'] = metadata.get('source')
    else:
        metadata['source'] = metadata['original_source']
    
    if 'chunk_length' not in metadata:
        metadata['chunk_length'] = chunk_length
    
    return metadata
